#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import json
import mock

from cinder import test
from cinder.volume import configuration as conf
from cinder.volume.targets import spdknvmf as spdknvmf_driver


BDEVS = [{
    "num_blocks": 4096000,
    "name": "Nvme0n1",
    "driver_specific": {
        "nvme": {
            "trid": {
                "trtype": "PCIe",
                "traddr": "0000:00:04.0"
            },
            "ns_data": {
                "id": 1
            },
            "pci_address": "0000:00:04.0",
            "vs": {
                "nvme_version": "1.1"
            },
            "ctrlr_data": {
                "firmware_revision": "1.0",
                "serial_number": "deadbeef",
                "oacs": {
                    "ns_manage": 0,
                    "security": 0,
                    "firmware": 0,
                    "format": 0
                },
                "vendor_id": "0x8086",
                "model_number": "QEMU NVMe Ctrl"
            },
            "csts": {
                "rdy": 1,
                "cfs": 0
            }
        }
    },
    "supported_io_types": {
        "reset": True,
        "nvme_admin": True,
        "unmap": False,
        "read": True,
        "write_zeroes": False,
        "write": True,
        "flush": True,
        "nvme_io": True
    },
    "claimed": False,
    "block_size": 512,
    "product_name": "NVMe disk",
    "aliases": ["Nvme0n1"]
}, {
    "num_blocks": 8192,
    "uuid": "70efd305-4e66-49bd-99ff-faeda5c3052d",
    "aliases": [
        "Nvme0n1p0"
    ],
    "driver_specific": {
        "lvol": {
            "base_bdev": "Nvme0n1",
            "lvol_store_uuid": "58b17014-d4a1-4f85-9761-093643ed18f1",
            "thin_provision": False
        }
    },
    "supported_io_types": {
        "reset": True,
        "nvme_admin": False,
        "unmap": True,
        "read": True,
        "write_zeroes": True,
        "write": True,
        "flush": False,
        "nvme_io": False
    },
    "claimed": False,
    "block_size": 4096,
    "product_name": "Split Disk",
    "name": "Nvme0n1p0"
}, {
    "num_blocks": 8192,
    "uuid": "70efd305-4e66-49bd-99ff-faeda5c3052d",
    "aliases": [
        "Nvme0n1p1"
    ],
    "driver_specific": {
        "lvol": {
            "base_bdev": "Nvme0n1",
            "lvol_store_uuid": "58b17014-d4a1-4f85-9761-093643ed18f1",
            "thin_provision": False
        }
    },
    "supported_io_types": {
        "reset": True,
        "nvme_admin": False,
        "unmap": True,
        "read": True,
        "write_zeroes": True,
        "write": True,
        "flush": False,
        "nvme_io": False
    },
    "claimed": False,
    "block_size": 4096,
    "product_name": "Split Disk",
    "name": "Nvme0n1p1"
}, {
    "num_blocks": 8192,
    "uuid": "70efd305-4e66-49bd-99ff-faeda5c3052d",
    "aliases": [
        "lvs_test/lvol0"
    ],
    "driver_specific": {
        "lvol": {
            "base_bdev": "Malloc0",
            "lvol_store_uuid": "58b17014-d4a1-4f85-9761-093643ed18f1",
            "thin_provision": False
        }
    },
    "supported_io_types": {
        "reset": True,
        "nvme_admin": False,
        "unmap": True,
        "read": True,
        "write_zeroes": True,
        "write": True,
        "flush": False,
        "nvme_io": False
    },
    "claimed": False,
    "block_size": 4096,
    "product_name": "Logical Volume",
    "name": "58b17014-d4a1-4f85-9761-093643ed18f1_4294967297"
}, {
    "num_blocks": 8192,
    "uuid": "8dec1964-d533-41df-bea7-40520efdb416",
    "aliases": [
        "lvs_test/lvol1"
    ],
    "driver_specific": {
        "lvol": {
            "base_bdev": "Malloc0",
            "lvol_store_uuid": "58b17014-d4a1-4f85-9761-093643ed18f1",
            "thin_provision": True
        }
    },
    "supported_io_types": {
        "reset": True,
        "nvme_admin": False,
        "unmap": True,
        "read": True,
        "write_zeroes": True,
        "write": True,
        "flush": False,
        "nvme_io": False
    },
    "claimed": False,
    "block_size": 4096,
    "product_name": "Logical Volume",
    "name": "58b17014-d4a1-4f85-9761-093643ed18f1_4294967298"
}]


NVMF_SUBSYSTEMS = [{
    "listen_addresses": [],
    "subtype": "Discovery",
    "nqn": "nqn.2014-08.org.nvmexpress.discovery",
    "hosts": [],
    "allow_any_host": True
}, {
    "listen_addresses": [],
    "subtype": "NVMe",
    "hosts": [{
        "nqn": "nqn.2016-06.io.spdk:init"
    }],
    "namespaces": [{
        "bdev_name": "Nvme0n1p0",
        "nsid": 1,
        "name": "Nvme0n1p0"
    }],
    "allow_any_host": False,
    "serial_number": "SPDK00000000000001",
    "nqn": "nqn.2016-06.io.spdk:cnode1"
}, {
    "listen_addresses": [],
    "subtype": "NVMe",
    "hosts": [],
    "namespaces": [{
        "bdev_name": "Nvme1n1p0",
        "nsid": 1,
        "name": "Nvme1n1p0"
    }],
    "allow_any_host": True,
    "serial_number": "SPDK00000000000002",
    "nqn": "nqn.2016-06.io.spdk:cnode2"
}]


class JSONRPCException(Exception):
    def __init__(self, message):
        self.message = message


class JSONRPCClient(object):
    def __init__(self, addr=None, port=None):
        self.methods = {"get_bdevs": self.get_bdevs,
                        "construct_nvmf_subsystem":
                            self.construct_nvmf_subsystem,
                        "delete_nvmf_subsystem": self.delete_nvmf_subsystem,
                        "nvmf_subsystem_create": self.nvmf_subsystem_create,
                        "nvmf_subsystem_add_listener":
                            self.nvmf_subsystem_add_listener,
                        "nvmf_subsystem_add_ns":
                            self.nvmf_subsystem_add_ns,
                        "get_nvmf_subsystems": self.get_nvmf_subsystems}
        self.bdevs = copy.deepcopy(BDEVS)
        self.nvmf_subsystems = copy.deepcopy(NVMF_SUBSYSTEMS)

    def __del__(self):
        pass

    def get_bdevs(self, params=None):
        if params and 'name' in params:
            for bdev in self.bdevs:
                for alias in bdev['aliases']:
                    if params['name'] in alias:
                        return json.dumps({"result": [bdev]})
                if bdev['name'] == params['name']:
                    return json.dumps({"result": [bdev]})
            return json.dumps({"error": "Not found"})

        return json.dumps({"result": self.bdevs})

    def get_nvmf_subsystems(self, params=None):
        return json.dumps({"result": self.nvmf_subsystems})

    def construct_nvmf_subsystem(self, params=None):
        nvmf_subsystem = {
            "listen_addresses": [],
            "subtype": "NVMe",
            "hosts": [],
            "namespaces": [{
                "bdev_name": "Nvme1n1p0",
                "nsid": 1,
                "name": "Nvme1n1p0"
            }],
            "allow_any_host": True,
            "serial_number": params['serial_number'],
            "nqn": params['nqn']
        }
        self.nvmf_subsystems.append(nvmf_subsystem)

        return json.dumps({"result": nvmf_subsystem})

    def delete_nvmf_subsystem(self, params=None):
        found_id = -1
        i = 0
        for nvmf_subsystem in self.nvmf_subsystems:
            if nvmf_subsystem['nqn'] == params['nqn']:
                found_id = i
            i += 1

        if found_id != -1:
            del self.nvmf_subsystems[found_id]

        return json.dumps({"result": {}})

    def nvmf_subsystem_create(self, params=None):
        nvmf_subsystem = {
            "namespaces": [],
            "nqn": params['nqn'],
            "serial_number": "S0000000000000000001",
            "allow_any_host": False,
            "subtype": "NVMe",
            "hosts": [],
            "listen_addresses": []
        }

        self.nvmf_subsystems.append(nvmf_subsystem)

        return json.dumps({"result": nvmf_subsystem})

    def nvmf_subsystem_add_listener(self, params=None):
        for nvmf_subsystem in self.nvmf_subsystems:
            if nvmf_subsystem['nqn'] == params['nqn']:
                nvmf_subsystem['listen_addresses'].append(
                    params['listen_address']
                )

        return json.dumps({"result": ""})

    def nvmf_subsystem_add_ns(self, params=None):
        for nvmf_subsystem in self.nvmf_subsystems:
            if nvmf_subsystem['nqn'] == params['nqn']:
                nvmf_subsystem['namespaces'].append(
                    params['namespace']
                )

        return json.dumps({"result": ""})

    def call(self, method, params=None):
        req = {}
        req['jsonrpc'] = '2.0'
        req['method'] = method
        req['id'] = 1
        if (params):
            req['params'] = params
        response = json.loads(self.methods[method](params))
        if not response:
            return {}

        if 'error' in response:
            msg = "\n".join(["Got JSON-RPC error response",
                             "request:",
                             json.dumps(req, indent=2),
                             "response:",
                             json.dumps(response['error'], indent=2)])
            raise JSONRPCException(msg)

        return response['result']


class Target(object):
    def __init__(self, name="Nvme0n1p0"):
        self.name = name


class SpdkNvmfDriverTestCase(test.TestCase):
    def setUp(self):
        super(SpdkNvmfDriverTestCase, self).setUp()
        self.configuration = mock.Mock(conf.Configuration)
        self.configuration.target_ip_address = '192.168.0.1'
        self.configuration.target_port = '4420'
        self.configuration.target_prefix = ""
        self.configuration.nvmet_port_id = "1"
        self.configuration.nvmet_ns_id = "fake_id"
        self.configuration.nvmet_subsystem_name = "nqn.2014-08.io.spdk"
        self.configuration.target_protocol = "nvmet_rdma"
        self.configuration.spdk_rpc_ip = "127.0.0.1"
        self.configuration.spdk_rpc_port = 8000
        self.driver = spdknvmf_driver.SpdkNvmf(configuration=
                                               self.configuration)
        self.jsonrpcclient = JSONRPCClient()

    def test__get_spdk_volume_name(self):
        with mock.patch.object(self.driver, "_rpc_call",
                               self.jsonrpcclient.call):
            bdevs = self.driver._rpc_call("get_bdevs")
            bdev_name = bdevs[0]['name']
            volume_name = self.driver._get_spdk_volume_name(bdev_name)
            self.assertEqual(bdev_name, volume_name)
            volume_name = self.driver._get_spdk_volume_name("fake")
            self.assertIsNone(volume_name)

    def test__get_nqn_with_volume_name(self):
        with mock.patch.object(self.driver, "_rpc_call",
                               self.jsonrpcclient.call):
            nqn = self.driver._get_nqn_with_volume_name("Nvme0n1p0")
            nqn_tmp = self.driver._rpc_call("get_nvmf_subsystems")[1]['nqn']
            self.assertEqual(nqn, nqn_tmp)
            nqn = self.driver._get_nqn_with_volume_name("fake")
            self.assertIsNone(nqn)

    def test__get_first_free_node(self):
        with mock.patch.object(self.driver, "_rpc_call",
                               self.jsonrpcclient.call):
            free_node = self.driver._get_first_free_node()
            self.assertEqual(3, free_node)

    def test_create_nvmeof_target(self):
        with mock.patch.object(self.driver, "_rpc_call",
                               self.jsonrpcclient.call):
            subsystems_first = self.driver._rpc_call("get_nvmf_subsystems")
            self.driver.create_nvmeof_target("Nvme0n1p1",
                                             "nqn.2016-06.io.spdk",
                                             "192.168.0.1",
                                             4420, "rdma", -1, -1, "")
            subsystems_last = self.driver._rpc_call("get_nvmf_subsystems")
            self.assertEqual(len(subsystems_first) + 1, len(subsystems_last))

    def test_delete_nvmeof_target(self):
        with mock.patch.object(self.driver, "_rpc_call",
                               self.jsonrpcclient.call):
            subsystems_first = self.driver._rpc_call("get_nvmf_subsystems")
            target = Target()
            self.driver.delete_nvmeof_target(target)
            subsystems_last = self.driver._rpc_call("get_nvmf_subsystems")
            self.assertEqual(len(subsystems_first) - 1, len(subsystems_last))
            target.name = "fake"
            self.driver.delete_nvmeof_target(target)
            self.assertEqual(len(subsystems_first) - 1, len(subsystems_last))



# Copyright (c) 2018 Huawei Technologies Co., Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import ddt

import mock
import os
import shutil
import tempfile

from six.moves import configparser

from cinder import test
from cinder.volume.drivers.fusionstorage import fs_conf


@ddt.ddt
class FusionStorageConfTestCase(test.TestCase):
    def setUp(self):
        super(FusionStorageConfTestCase, self).setUp()
        self.tmp_dir = tempfile.mkdtemp()
        self.conf = mock.Mock()
        self._create_fake_conf_file()
        self.fusionstorage_conf = fs_conf.FusionStorageConf(
            self.conf, "cinder@fs")

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        super(FusionStorageConfTestCase, self).tearDown()

    def _create_fake_conf_file(self):
        self.conf.cinder_fusionstorage_conf_file = (
            self.tmp_dir + '/cinder.conf')

        config = configparser.ConfigParser()
        config.add_section('backend_name')
        config.set('backend_name', 'dsware_rest_url', 'https://fake_rest_site')
        config.set('backend_name', 'san_login', 'fake_user')
        config.set('backend_name', 'san_password', 'fake_passwd')
        config.set('backend_name', 'dsware_storage_pools', 'fake_pool')

        config.add_section('manager_ip')
        config.set('manager_ip', 'fake_host', 'fake_ip')
        config.write(open(self.conf.cinder_fusionstorage_conf_file, 'w'))

    @mock.patch.object(fs_conf.FusionStorageConf, '_encode_authentication')
    @mock.patch.object(fs_conf.FusionStorageConf, '_pools_name')
    @mock.patch.object(fs_conf.FusionStorageConf, '_san_address')
    @mock.patch.object(fs_conf.FusionStorageConf, '_san_user')
    @mock.patch.object(fs_conf.FusionStorageConf, '_san_password')
    def test_update_config_value(self, mock_san_password, mock_san_user,
                                 mock_san_address, mock_pools_name,
                                 mock_encode_authentication):
        self.fusionstorage_conf.update_config_value()
        mock_encode_authentication.assert_called_once_with()
        mock_pools_name.assert_called_once_with()
        mock_san_address.assert_called_once_with()
        mock_san_user.assert_called_once_with()
        mock_san_password.assert_called_once_with()

    @mock.patch.object(os.path, 'exists')
    def test__encode_authentication(self, mock_exists):
        config = configparser.ConfigParser()
        config.read(self.conf.cinder_fusionstorage_conf_file)
        mock_exists.return_value = False

        user_name = 'fake_user'
        self.mock_object(
            self.fusionstorage_conf.configuration, 'safe_get',
            return_value=user_name)
        self.fusionstorage_conf._encode_authentication()

        password = 'fake_passwd'
        self.mock_object(
            self.fusionstorage_conf.configuration, 'safe_get',
            return_value=password)
        self.fusionstorage_conf._encode_authentication()

    @mock.patch.object(os.path, 'exists')
    @mock.patch.object(configparser.ConfigParser, 'set')
    def test__rewrite_conf(self, mock_set, mock_exists):
        mock_exists.return_value = False
        mock_set.return_value = "success"
        self.fusionstorage_conf._rewrite_conf('fake_name', 'fake_pwd')

    def test__san_address(self):
        address = 'https://fake_rest_site'
        self.mock_object(
            self.fusionstorage_conf.configuration, 'safe_get',
            return_value=address)
        self.fusionstorage_conf._san_address()
        self.assertEqual('https://fake_rest_site',
                         self.fusionstorage_conf.configuration.san_address)

    def test__san_user(self):
        user = '!&&&ZmFrZV91c2Vy'
        self.mock_object(
            self.fusionstorage_conf.configuration, 'safe_get',
            return_value=user)
        self.fusionstorage_conf._san_user()
        self.assertEqual(
            'fake_user', self.fusionstorage_conf.configuration.san_user)

        user = 'fake_user_2'
        self.mock_object(
            self.fusionstorage_conf.configuration, 'safe_get',
            return_value=user)
        self.fusionstorage_conf._san_user()
        self.assertEqual(
            'fake_user_2', self.fusionstorage_conf.configuration.san_user)

    def test__san_password(self):
        password = '!&&&ZmFrZV9wYXNzd2Q='
        self.mock_object(
            self.fusionstorage_conf.configuration, 'safe_get',
            return_value=password)
        self.fusionstorage_conf._san_password()
        self.assertEqual(
            'fake_passwd', self.fusionstorage_conf.configuration.san_password)

        password = 'fake_passwd_2'
        self.mock_object(
            self.fusionstorage_conf.configuration, 'safe_get',
            return_value=password)
        self.fusionstorage_conf._san_password()
        self.assertEqual('fake_passwd_2',
                         self.fusionstorage_conf.configuration.san_password)

    def test__pools_name(self):
        pools_name = 'fake_pool'
        self.mock_object(
            self.fusionstorage_conf.configuration, 'safe_get',
            return_value=pools_name)
        self.fusionstorage_conf._pools_name()
        self.assertListEqual(
            ['fake_pool'], self.fusionstorage_conf.configuration.pools_name)

    def test__manager_ip(self):
        manager_ips = {'fake_host': 'fake_ip'}
        self.mock_object(
            self.fusionstorage_conf.configuration, 'safe_get',
            return_value=manager_ips)
        self.fusionstorage_conf._manager_ip()
        self.assertDictEqual({'fake_host': 'fake_ip'},
                             self.fusionstorage_conf.configuration.manager_ips)



#    Copyright (c) 2015-2017 Dell Inc, or its subsidiaries.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Volume driver for Dell Storage Center."""

from oslo_log import log as logging
from oslo_utils import excutils

from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder import utils
from cinder.volume import driver
from cinder.volume.drivers.dell_emc.sc import storagecenter_common
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)


@interface.volumedriver
class SCFCDriver(storagecenter_common.SCCommonDriver,
                 driver.FibreChannelDriver):

    """Implements commands for Dell Storage Center FC management.

    To enable the driver add the following line to the cinder configuration:
        volume_driver=cinder.volume.drivers.dell_emc.sc.storagecenter_fc.\
        SCFCDriver

    Version history:

    .. code-block:: none

        1.0.0 - Initial driver
        1.1.0 - Added extra spec support for Storage Profile selection
        1.2.0 - Added consistency group support.
        2.0.0 - Switched to inheriting functional objects rather than volume
                driver.
        2.1.0 - Added support for ManageableVD.
        2.2.0 - Driver retype support for switching volume's Storage Profile
        2.3.0 - Added Legacy Port Mode Support
        2.3.1 - Updated error handling.
        2.4.0 - Added Replication V2 support.
        2.4.1 - Updated Replication support to V2.1.
        2.5.0 - ManageableSnapshotsVD implemented.
        3.0.0 - ProviderID utilized.
        3.1.0 - Failback supported.
        3.2.0 - Live Volume support.
        3.3.0 - Support for a secondary DSM.
        3.4.0 - Support for excluding a domain.
        3.5.0 - Support for AFO.
        3.6.0 - Server type support.
        3.7.0 - Support for Data Reduction, Group QOS and Volume QOS.
        4.0.0 - Driver moved to dell_emc.
        4.1.0 - Timeouts added to rest calls.
        4.1.1 - excluded_domain_ips support.

    """

    VERSION = '4.1.1'

    CI_WIKI_NAME = "Dell_EMC_SC_Series_CI"

    def __init__(self, *args, **kwargs):
        super(SCFCDriver, self).__init__(*args, **kwargs)
        self.backend_name =\
            self.configuration.safe_get('volume_backend_name') or 'Dell-FC'
        self.storage_protocol = 'FC'

    def validate_connector(self, connector):
        """Fail if connector doesn't contain all the data needed by driver.

        Do a check on the connector and ensure that it has wwnns, wwpns.
        """
        self.validate_connector_has_setting(connector, 'wwpns')
        self.validate_connector_has_setting(connector, 'wwnns')

    def initialize_connection(self, volume, connector):
        """Initializes the connection and returns connection info.

        Assign any created volume to a compute node/host so that it can be
        used from that host.

        The  driver returns a driver_volume_type of 'fibre_channel'.
        The target_wwn can be a single entry or a list of wwns that
        correspond to the list of remote wwn(s) that will export the volume.
        """

        # We use id to name the volume name as it is a
        # known unique name.
        volume_name = volume.get('id')
        provider_id = volume.get('provider_id')
        islivevol = self._is_live_vol(volume)
        LOG.debug('Initialize connection: %s', volume_name)
        with self._client.open_connection() as api:
            try:
                wwpns = connector.get('wwpns')
                # Find the volume on the storage center. Note that if this
                # is live volume and we are swapped this will be the back
                # half of the live volume.
                scvolume = api.find_volume(volume_name, provider_id, islivevol)
                if scvolume:
                    # Get the SSN it is on.
                    ssn = scvolume['instanceId'].split('.')[0]
                    # Find our server.
                    scserver = self._find_server(api, wwpns, ssn)

                    # No? Create it.
                    if scserver is None:
                        scserver = api.create_server(
                            wwpns, self.configuration.dell_server_os, ssn)
                    # We have a volume and a server. Map them.
                    if scserver is not None:
                        mapping = api.map_volume(scvolume, scserver)
                        if mapping is not None:
                            # Since we just mapped our volume we had
                            # best update our sc volume object.
                            scvolume = api.get_volume(scvolume['instanceId'])
                            lun, targets, init_targ_map = api.find_wwns(
                                scvolume, scserver)

                            # Do we have extra live volume work?
                            if islivevol:
                                # Get our live volume.
                                sclivevolume = api.get_live_volume(provider_id)
                                # Do not map to a failed over volume.
                                if (sclivevolume and not
                                    api.is_failed_over(provider_id,
                                                       sclivevolume)):
                                    # Now map our secondary.
                                    lvlun, lvtargets, lvinit_targ_map = (
                                        self.initialize_secondary(api,
                                                                  sclivevolume,
                                                                  wwpns))
                                    # Unmapped. Add info to our list.
                                    targets += lvtargets
                                    init_targ_map.update(lvinit_targ_map)

                            # Roll up our return data.
                            if lun is not None and len(targets) > 0:
                                data = {'driver_volume_type': 'fibre_channel',
                                        'data': {'target_lun': lun,
                                                 'target_discovered': True,
                                                 'target_wwn': targets,
                                                 'initiator_target_map':
                                                 init_targ_map,
                                                 'discard': True}}
                                LOG.debug('Return FC data: %s', data)
                                fczm_utils.add_fc_zone(data)
                                return data
                            LOG.error('Lun mapping returned null!')

            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error('Failed to initialize connection.')

        # We get here because our mapping is none so blow up.
        raise exception.VolumeBackendAPIException(
            data=_('Unable to map volume.'))

    def _find_server(self, api, wwns, ssn=-1):
        for wwn in wwns:
            scserver = api.find_server(wwn, ssn)
            if scserver is not None:
                return scserver
        return None

    def initialize_secondary(self, api, sclivevolume, wwns):
        """Initialize the secondary connection of a live volume pair.

        :param api: Dell SC api object.
        :param sclivevolume: Dell SC live volume object.
        :param wwns: Cinder list of wwns from the connector.
        :return: lun, targets and initiator target map.
        """
        # Find our server.
        secondary = self._find_server(
            api, wwns, sclivevolume['secondaryScSerialNumber'])

        # No? Create it.
        if secondary is None:
            secondary = api.create_server(
                wwns, self.configuration.dell_server_os,
                sclivevolume['secondaryScSerialNumber'])
        if secondary:
            if api.map_secondary_volume(sclivevolume, secondary):
                # Get mappings.
                secondaryvol = api.get_volume(
                    sclivevolume['secondaryVolume']['instanceId'])
                if secondaryvol:
                    return api.find_wwns(secondaryvol, secondary)
        LOG.warning('Unable to map live volume secondary volume'
                    ' %(vol)s to secondary server wwns: %(wwns)r',
                    {'vol': sclivevolume['secondaryVolume']['instanceName'],
                     'wwns': wwns})
        return None, [], {}

    def force_detach(self, volume):
        """Breaks all volume server connections including to the live volume.

        :param volume: volume to be detached
        :raises VolumeBackendAPIException: On failure to sever connections.
        """
        with self._client.open_connection() as api:
            volume_name = volume.get('id')
            provider_id = volume.get('provider_id')
            try:
                islivevol = self._is_live_vol(volume)
                scvolume = api.find_volume(volume_name, provider_id, islivevol)
                if scvolume:
                    rtn = api.unmap_all(scvolume)
                    # If this fails we blow up.
                    if not rtn:
                        raise exception.VolumeBackendAPIException(
                            _('Terminate connection failed'))
                    # If there is a livevol we just take a shot at
                    # disconnecting.
                    if islivevol:
                        sclivevolume = api.get_live_volume(provider_id)
                        if sclivevolume:
                            self.terminate_secondary(api, sclivevolume, None)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error('Failed to terminates %(vol)s connections.',
                              {'vol': volume_name})

        # We don't know the servers that were involved so we just return
        # the most basic of data.
        info = {'driver_volume_type': 'fibre_channel',
                'data': {}}
        return info

    @utils.synchronized('{self.driver_prefix}-{volume.id}')
    def terminate_connection(self, volume, connector, force=False, **kwargs):
        # Special case
        if connector is None:
            return self.force_detach(volume)

        # Grab some quick info.
        volume_name = volume.get('id')
        provider_id = volume.get('provider_id')
        LOG.debug('Terminate connection: %s', volume_name)
        LOG.debug('Volume details %s', volume)

        # None `connector` indicates force detach, then detach all even if the
        # volume is multi-attached.
        is_multiattached = (hasattr(volume, 'volume_attachment') and
                            self.is_multiattach_to_host(
                                volume.get('volume_attachment'),
                                connector['host']))
        if is_multiattached:
            LOG.info('Cannot terminate connection: '
                     '%(vol)s is multiattached.',
                     {'vol': volume_name})

            return True

        with self._client.open_connection() as api:
            try:
                wwpns = [] if not connector else connector.get('wwpns', [])
                # Find the volume on the storage center.
                islivevol = self._is_live_vol(volume)
                scvolume = api.find_volume(volume_name, provider_id, islivevol)
                if scvolume:
                    # Get the SSN it is on.
                    ssn = scvolume['instanceId'].split('.')[0]

                    # Will be None if we have no wwpns.
                    scserver = self._find_server(api, wwpns, ssn)

                    # Get our target map so we can return it to free up a zone.
                    lun, targets, init_targ_map = api.find_wwns(scvolume,
                                                                scserver)

                    # Do we have extra live volume work?
                    if islivevol:
                        # Get our live volume.
                        sclivevolume = api.get_live_volume(provider_id)
                        # Do not map to a failed over volume.
                        if (sclivevolume and not
                            api.is_failed_over(provider_id,
                                               sclivevolume)):
                            lvlun, lvtargets, lvinit_targ_map = (
                                self.terminate_secondary(
                                    api, sclivevolume, wwpns))
                            # Add to our return.
                            if lvlun:
                                targets += lvtargets
                                init_targ_map.update(lvinit_targ_map)

                    if (wwpns and scserver and
                            api.unmap_volume(scvolume, scserver) is True):
                        LOG.debug('Connection terminated')
                    elif not wwpns and api.unmap_all(scvolume):
                        LOG.debug('All connections terminated')
                    else:
                        raise exception.VolumeBackendAPIException(
                            data=_('Terminate connection failed'))

                    info = {'driver_volume_type': 'fibre_channel',
                            'data': {}}

                    # if not then we return the target map so that
                    # the zone can be freed up.
                    if scserver and api.get_volume_count(scserver) == 0:
                        info['data'] = {'target_wwn': targets,
                                        'initiator_target_map': init_targ_map}
                        fczm_utils.remove_fc_zone(info)
                    return info

            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error('Failed to terminate connection')
        raise exception.VolumeBackendAPIException(
            data=_('Terminate connection unable to connect to backend.'))

    def terminate_secondary(self, api, sclivevolume, wwns):
        lun = None
        targets = []
        init_targ_map = {}
        # Get our volume.
        secondaryvol = api.get_volume(
            sclivevolume['secondaryVolume']['instanceId'])
        # We have one so let's get to work.
        if secondaryvol:
            # Are we unmapping a specific server?
            if wwns:
                # Find our server.
                secondary = self._find_server(
                    api, wwns, sclivevolume['secondaryScSerialNumber'])
                # Get our map.
                lun, targets, init_targ_map = api.find_wwns(secondaryvol,
                                                            secondary)
                # If we have a server and a volume lets unmap them.
                ret = api.unmap_volume(secondaryvol, secondary)
                LOG.debug('terminate_secondary: '
                          'secondary volume %(name)s unmap '
                          'to secondary server %(server)s result: %(result)r',
                          {'name': secondaryvol['name'],
                           'server': secondary['name'], 'result': ret})
            else:
                # Just unmap all.
                ret = api.unmap_all(secondaryvol)
                LOG.debug('terminate_secondary:  secondary volume %(name)s '
                          'unmap all result: %(result)r',
                          {'name': secondaryvol['name'], 'result': ret})
        else:
            LOG.debug('terminate_secondary: secondary volume not found.')
        # return info if any
        return lun, targets, init_targ_map



# Copyright (c) 2017 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


class PQueue(object):

    def __init__(self, path, interval=None):
        self.path = path
        self._interval = interval
        self.started = False

    def put(self, item):
        return item

    def start(self):
        self.started = True



# Copyright 2016 Dell Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

_volume_register = []
_backup_register = []
_fczm_register = []


def volumedriver(cls):
    """Decorator for concrete volume driver implementations."""
    _volume_register.append(cls)
    return cls


def backupdriver(cls):
    """Decorator for concrete backup driver implementations."""
    _backup_register.append(cls)
    return cls


def fczmdriver(cls):
    """Decorator for concrete fibre channel zone manager drivers."""
    _fczm_register.append(cls)
    return cls



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock
from oslo_config import cfg

from cinder import context
from cinder import db
from cinder import exception
from cinder import objects
from cinder.objects import fields
from cinder import service
from cinder.tests.unit.api import fakes
from cinder.tests.unit import utils as tests_utils
from cinder.tests.unit import volume as base


CONF = cfg.CONF


class VolumeCleanupTestCase(base.BaseVolumeTestCase):
    MOCK_WORKER = False

    def setUp(self):
        super(VolumeCleanupTestCase, self).setUp()
        self.service_id = 1
        self.mock_object(service.Service, 'service_id', self.service_id)
        self.patch('cinder.volume.volume_utils.clear_volume', autospec=True)

    def _assert_workers_are_removed(self):
        workers = db.worker_get_all(self.context, read_deleted='yes')
        self.assertListEqual([], workers)

    def test_init_host_clears_uploads_available_volume(self):
        """init_host will clean an available volume stuck in uploading."""
        volume = tests_utils.create_volume(self.context, status='uploading',
                                           size=0, host=CONF.host)

        db.worker_create(self.context, resource_type='Volume',
                         resource_id=volume.id, status=volume.status,
                         service_id=self.service_id)

        self.volume.init_host(service_id=service.Service.service_id)
        volume.refresh()
        self.assertEqual("available", volume.status)
        self._assert_workers_are_removed()

    @mock.patch('cinder.manager.CleanableManager.init_host')
    def test_init_host_clears_uploads_in_use_volume(self, init_host_mock):
        """init_host will clean an in-use volume stuck in uploading."""
        volume = tests_utils.create_volume(self.context, status='uploading',
                                           size=0, host=CONF.host)

        db.worker_create(self.context, resource_type='Volume',
                         resource_id=volume.id, status=volume.status,
                         service_id=self.service_id)

        fake_uuid = fakes.get_fake_uuid()
        tests_utils.attach_volume(self.context, volume.id, fake_uuid,
                                  'fake_host', '/dev/vda')
        self.volume.init_host(service_id=mock.sentinel.service_id)
        init_host_mock.assert_called_once_with(
            service_id=mock.sentinel.service_id, added_to_cluster=None)
        volume.refresh()
        self.assertEqual("in-use", volume.status)
        self._assert_workers_are_removed()

    @mock.patch('cinder.image.image_utils.cleanup_temporary_file')
    def test_init_host_clears_downloads(self, mock_cleanup_tmp_file):
        """Test that init_host will unwedge a volume stuck in downloading."""
        volume = tests_utils.create_volume(self.context, status='downloading',
                                           size=0, host=CONF.host)
        db.worker_create(self.context, resource_type='Volume',
                         resource_id=volume.id, status=volume.status,
                         service_id=self.service_id)
        mock_clear = self.mock_object(self.volume.driver, 'clear_download')

        self.volume.init_host(service_id=service.Service.service_id)
        self.assertEqual(1, mock_clear.call_count)
        self.assertEqual(volume.id, mock_clear.call_args[0][1].id)
        volume.refresh()
        self.assertEqual("error", volume['status'])
        mock_cleanup_tmp_file.assert_called_once_with(CONF.host)

        self.volume.delete_volume(self.context, volume=volume)
        self._assert_workers_are_removed()

    @mock.patch('cinder.image.image_utils.cleanup_temporary_file')
    def test_init_host_resumes_deletes(self, mock_cleanup_tmp_file):
        """init_host will resume deleting volume in deleting status."""
        volume = tests_utils.create_volume(self.context, status='deleting',
                                           size=0, host=CONF.host)

        db.worker_create(self.context, resource_type='Volume',
                         resource_id=volume.id, status=volume.status,
                         service_id=self.service_id)

        self.volume.init_host(service_id=service.Service.service_id)

        self.assertRaises(exception.VolumeNotFound, db.volume_get,
                          context.get_admin_context(), volume.id)
        mock_cleanup_tmp_file.assert_called_once_with(CONF.host)
        self._assert_workers_are_removed()

    @mock.patch('cinder.image.image_utils.cleanup_temporary_file')
    def test_create_volume_fails_with_creating_and_downloading_status(
            self, mock_cleanup_tmp_file):
        """Test init_host_with_service in case of volume.

        While the status of volume is 'creating' or 'downloading',
        volume process down.
        After process restarting this 'creating' status is changed to 'error'.
        """
        for status in ('creating', 'downloading'):
            volume = tests_utils.create_volume(self.context, status=status,
                                               size=0, host=CONF.host)

            db.worker_create(self.context, resource_type='Volume',
                             resource_id=volume.id, status=volume.status,
                             service_id=self.service_id)

            self.volume.init_host(service_id=service.Service.service_id)
            volume.refresh()

            self.assertEqual('error', volume['status'])
            self.volume.delete_volume(self.context, volume)
            self.assertTrue(mock_cleanup_tmp_file.called)
            self._assert_workers_are_removed()

    def test_create_snapshot_fails_with_creating_status(self):
        """Test init_host_with_service in case of snapshot.

        While the status of snapshot is 'creating', volume process
        down. After process restarting this 'creating' status is
        changed to 'error'.
        """
        volume = tests_utils.create_volume(self.context,
                                           **self.volume_params)
        snapshot = tests_utils.create_snapshot(
            self.context,
            volume.id,
            status=fields.SnapshotStatus.CREATING)
        db.worker_create(self.context, resource_type='Snapshot',
                         resource_id=snapshot.id, status=snapshot.status,
                         service_id=self.service_id)

        self.volume.init_host(service_id=service.Service.service_id)

        snapshot_obj = objects.Snapshot.get_by_id(self.context, snapshot.id)

        self.assertEqual(fields.SnapshotStatus.ERROR, snapshot_obj.status)
        self.assertEqual(service.Service.service_id,
                         self.volume.service_id)
        self._assert_workers_are_removed()

        self.volume.delete_snapshot(self.context, snapshot_obj)
        self.volume.delete_volume(self.context, volume)

    def test_init_host_clears_deleting_snapshots(self):
        """Test that init_host will delete a snapshot stuck in deleting."""
        volume = tests_utils.create_volume(self.context, status='deleting',
                                           size=1, host=CONF.host)
        snapshot = tests_utils.create_snapshot(self.context,
                                               volume.id, status='deleting')

        db.worker_create(self.context, resource_type='Volume',
                         resource_id=volume.id, status=volume.status,
                         service_id=self.service_id)

        self.volume.init_host(service_id=self.service_id)
        self.assertRaises(exception.VolumeNotFound, volume.refresh)
        self.assertRaises(exception.SnapshotNotFound, snapshot.refresh)



#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils.uuidutils import is_uuid_like
from oslo_versionedobjects import fields

from cinder.db.sqlalchemy import models
from cinder import objects
from cinder.objects import fields as c_fields
from cinder.tests.unit import fake_constants as fake


def fake_db_volume(**updates):
    db_volume = {
        'id': fake.VOLUME_ID,
        'size': 1,
        'name': 'volume-%s' % fake.VOLUME_ID,
        'availability_zone': 'fake_availability_zone',
        'status': 'available',
        'attach_status': c_fields.VolumeAttachStatus.DETACHED,
        'previous_status': None,
        'volume_attachment': [],
        'volume_metadata': [],
        'volume_admin_metadata': [],
        'volume_glance_metadata': [],
        'snapshots': [],
    }

    for name, field in objects.Volume.fields.items():
        if name in db_volume:
            continue
        if field.nullable:
            db_volume[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_volume[name] = field.default
        else:
            raise Exception('fake_db_volume needs help with %s.' % name)

    if updates:
        db_volume.update(updates)

    return db_volume


def fake_db_volume_type(**updates):
    db_volume_type = {
        'id': fake.VOLUME_TYPE_ID,
        'name': 'type-1',
        'description': 'A fake volume type',
        'is_public': True,
        'projects': [],
        'extra_specs': {},
    }

    for name, field in objects.VolumeType.fields.items():
        if name in db_volume_type:
            continue
        if field.nullable:
            db_volume_type[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_volume_type[name] = field.default
        else:
            raise Exception('fake_db_volume_type needs help with %s.' % name)

    if updates:
        db_volume_type.update(updates)

    return db_volume_type


def fake_db_volume_attachment(**updates):
    db_volume_attachment = {
        'id': fake.ATTACHMENT_ID,
        'volume_id': fake.VOLUME_ID,
        'volume': fake_db_volume(),
    }

    for name, field in objects.VolumeAttachment.fields.items():
        if name in db_volume_attachment:
            continue
        if field.nullable:
            db_volume_attachment[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_volume_attachment[name] = field.default
        else:
            raise Exception(
                'fake_db_volume_attachment needs help with %s.' % name)

    if updates:
        db_volume_attachment.update(updates)

    return db_volume_attachment


def fake_volume_obj(context, **updates):
    if updates.get('encryption_key_id'):
        assert is_uuid_like(updates['encryption_key_id'])

    expected_attrs = updates.pop('expected_attrs',
                                 ['metadata', 'admin_metadata'])
    vol = objects.Volume._from_db_object(context, objects.Volume(),
                                         fake_db_volume(**updates),
                                         expected_attrs=expected_attrs)
    return vol


def fake_volume_type_obj(context, **updates):
    return objects.VolumeType._from_db_object(
        context, objects.VolumeType(), fake_db_volume_type(**updates))


def fake_volume_attachment_obj(context, **updates):
    return objects.VolumeAttachment._from_db_object(
        context, objects.VolumeAttachment(),
        fake_db_volume_attachment(**updates))


def volume_db_obj(**updates):
    """Return a volume ORM object."""
    updates.setdefault('id', fake.VOLUME_ID)
    updates.setdefault('size', 1)
    return models.Volume(**updates)


def volume_attachment_db_obj(**updates):
    updates.setdefault('id', fake.ATTACHMENT_ID)
    updates.setdefault('volume_id', fake.VOLUME_ID)
    updates.setdefault('volume', volume_db_obj())
    return models.VolumeAttachment(**updates)


def volume_attachment_ovo(context, **updates):
    orm = volume_attachment_db_obj(**updates)
    return objects.VolumeAttachment._from_db_object(context,
                                                    objects.VolumeAttachment(),
                                                    orm)



# Copyright 2019 Nexenta Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Unit tests for OpenStack Cinder volume driver
"""
import uuid

import mock
from oslo_utils import units

from cinder import context
from cinder import db
from cinder import test
from cinder.tests.unit.consistencygroup.fake_cgsnapshot import (
    fake_cgsnapshot_obj as fake_cgsnapshot)
from cinder.tests.unit.consistencygroup.fake_consistencygroup import (
    fake_consistencyobject_obj as fake_cgroup)
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit.fake_snapshot import fake_snapshot_obj as fake_snapshot
from cinder.tests.unit.fake_volume import fake_volume_obj as fake_volume
from cinder.volume import configuration as conf
from cinder.volume.drivers.nexenta.ns5 import iscsi
from cinder.volume.drivers.nexenta.ns5 import jsonrpc


class TestNexentaISCSIDriver(test.TestCase):

    def setUp(self):
        super(TestNexentaISCSIDriver, self).setUp()
        self.ctxt = context.get_admin_context()
        self.cfg = mock.Mock(spec=conf.Configuration)
        self.cfg.volume_backend_name = 'nexenta_iscsi'
        self.cfg.nexenta_group_snapshot_template = 'group-snapshot-%s'
        self.cfg.nexenta_origin_snapshot_template = 'origin-snapshot-%s'
        self.cfg.nexenta_dataset_description = ''
        self.cfg.nexenta_host = '1.1.1.1'
        self.cfg.nexenta_user = 'admin'
        self.cfg.nexenta_password = 'nexenta'
        self.cfg.nexenta_volume = 'cinder'
        self.cfg.nexenta_rest_port = 8443
        self.cfg.nexenta_use_https = False
        self.cfg.nexenta_iscsi_target_portal_port = 3260
        self.cfg.nexenta_target_prefix = 'iqn:cinder'
        self.cfg.nexenta_target_group_prefix = 'cinder'
        self.cfg.nexenta_ns5_blocksize = 32
        self.cfg.nexenta_sparse = True
        self.cfg.nexenta_lu_writebackcache_disabled = True
        self.cfg.nexenta_dataset_compression = 'on'
        self.cfg.nexenta_dataset_dedup = 'off'
        self.cfg.reserved_percentage = 20
        self.cfg.nexenta_host_group_prefix = 'hg'
        self.cfg.nexenta_volume = 'pool'
        self.cfg.driver_ssl_cert_verify = False
        self.cfg.nexenta_luns_per_target = 20
        self.cfg.driver_ssl_cert_verify = False
        self.cfg.nexenta_iscsi_target_portals = '1.1.1.1:3260,2.2.2.2:3260'
        self.cfg.nexenta_iscsi_target_host_group = 'all'
        self.cfg.nexenta_rest_address = '1.1.1.1'
        self.cfg.nexenta_rest_backoff_factor = 1
        self.cfg.nexenta_rest_retry_count = 3
        self.cfg.nexenta_rest_connect_timeout = 1
        self.cfg.nexenta_rest_read_timeout = 1
        self.cfg.nexenta_volume_group = 'vg'
        self.cfg.safe_get = self.fake_safe_get
        self.nef_mock = mock.Mock()
        self.mock_object(jsonrpc, 'NefRequest',
                         return_value=self.nef_mock)
        self.drv = iscsi.NexentaISCSIDriver(
            configuration=self.cfg)
        self.drv.db = db
        self.drv.do_setup(self.ctxt)

    def fake_safe_get(self, key):
        try:
            value = getattr(self.cfg, key)
        except AttributeError:
            value = None
        return value

    def fake_uuid4():
        return uuid.UUID('38d18a48-b791-4046-b523-a84aad966310')

    def test_do_setup(self):
        self.assertIsNone(self.drv.do_setup(self.ctxt))

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefServices.get')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumeGroups.create')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumeGroups.get')
    def test_check_for_setup_error(self, volume_group_get,
                                   volume_group_create,
                                   service_get):
        path = self.drv.root_path
        bs = self.cfg.nexenta_ns5_blocksize * units.Ki
        name = 'iscsit'
        state = 'online'
        volume_group_get.return_value = {'path': path}
        service_get.return_value = {'name': name, 'state': state}
        self.assertIsNone(self.drv.check_for_setup_error())
        volume_group_get.assert_called_with(path)
        service_get.assert_called_with(name)

        volume_group_get.side_effect = jsonrpc.NefException({
            'message': 'Failed to open dataset',
            'code': 'ENOENT'
        })
        volume_group_create.return_value = {}
        self.assertIsNone(self.drv.check_for_setup_error())
        volume_group_get.assert_called_with(path)
        payload = {'path': path, 'volumeBlockSize': bs}
        volume_group_create.assert_called_with(payload)
        service_get.assert_called_with(name)

        state = 'offline'
        volume_group_get.return_value = {'path': path}
        service_get.return_value = {'name': name, 'state': state}
        self.assertRaises(jsonrpc.NefException,
                          self.drv.check_for_setup_error)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumes.create')
    def test_create_volume(self, create_volume):
        volume = fake_volume(self.ctxt)
        self.assertIsNone(self.drv.create_volume(volume))
        path = self.drv._get_volume_path(volume)
        size = volume['size'] * units.Gi
        bs = self.cfg.nexenta_ns5_blocksize * units.Ki
        payload = {
            'path': path,
            'volumeSize': size,
            'volumeBlockSize': bs,
            'compressionMode': self.cfg.nexenta_dataset_compression,
            'sparseVolume': self.cfg.nexenta_sparse
        }
        create_volume.assert_called_with(payload)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumes.delete')
    def test_delete_volume(self, delete_volume):
        volume = fake_volume(self.ctxt)
        self.assertIsNone(self.drv.delete_volume(volume))
        path = self.drv._get_volume_path(volume)
        payload = {'snapshots': True}
        delete_volume.assert_called_with(path, payload)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumes.set')
    def test_extend_volume(self, extend_volume):
        volume = fake_volume(self.ctxt)
        size = volume['size'] * 2
        self.assertIsNone(self.drv.extend_volume(volume, size))
        path = self.drv._get_volume_path(volume)
        size = size * units.Gi
        payload = {'volumeSize': size}
        extend_volume.assert_called_with(path, payload)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.delete')
    def test_delete_snapshot(self, delete_snapshot):
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        delete_snapshot.return_value = {}
        self.assertIsNone(self.drv.delete_snapshot(snapshot))
        path = self.drv._get_snapshot_path(snapshot)
        payload = {'defer': True}
        delete_snapshot.assert_called_with(path, payload)

    def test_snapshot_revert_use_temp_snapshot(self):
        result = self.drv.snapshot_revert_use_temp_snapshot()
        expected = False
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumes.rollback')
    def test_revert_to_snapshot(self, rollback_volume):
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        rollback_volume.return_value = {}
        self.assertIsNone(
            self.drv.revert_to_snapshot(self.ctxt, volume, snapshot)
        )
        path = self.drv._get_volume_path(volume)
        payload = {'snapshot': snapshot['name']}
        rollback_volume.assert_called_with(path, payload)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.iscsi.'
                'NexentaISCSIDriver.delete_snapshot')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.iscsi.'
                'NexentaISCSIDriver.create_volume_from_snapshot')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.iscsi.'
                'NexentaISCSIDriver.create_snapshot')
    def test_create_cloned_volume(self, create_snapshot, create_volume,
                                  delete_snapshot):
        volume = fake_volume(self.ctxt)
        clone_spec = {'id': fake.VOLUME2_ID}
        clone = fake_volume(self.ctxt, **clone_spec)
        create_snapshot.return_value = {}
        create_volume.return_value = {}
        delete_snapshot.return_value = {}
        self.assertIsNone(self.drv.create_cloned_volume(clone, volume))
        snapshot = {
            'name': self.drv.origin_snapshot_template % clone['id'],
            'volume_id': volume['id'],
            'volume_name': volume['name'],
            'volume_size': volume['size']
        }
        create_snapshot.assert_called_with(snapshot)
        create_volume.assert_called_with(clone, snapshot)
        create_volume.side_effect = jsonrpc.NefException({
            'message': 'Failed to create volume',
            'code': 'EBUSY'
        })
        self.assertRaises(jsonrpc.NefException,
                          self.drv.create_cloned_volume,
                          clone, volume)
        create_snapshot.side_effect = jsonrpc.NefException({
            'message': 'Failed to open dataset',
            'code': 'ENOENT'
        })
        self.assertRaises(jsonrpc.NefException,
                          self.drv.create_cloned_volume,
                          clone, volume)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.create')
    def test_create_snapshot(self, create_snapshot):
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        create_snapshot.return_value = {}
        self.assertIsNone(self.drv.create_snapshot(snapshot))
        path = self.drv._get_snapshot_path(snapshot)
        payload = {'path': path}
        create_snapshot.assert_called_with(payload)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver.extend_volume')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.clone')
    def test_create_volume_from_snapshot(self, clone_snapshot,
                                         extend_volume):
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        clone_size = 10
        clone_spec = {
            'id': fake.VOLUME2_ID,
            'size': clone_size
        }
        clone = fake_volume(self.ctxt, **clone_spec)
        snapshot_path = self.drv._get_snapshot_path(snapshot)
        clone_path = self.drv._get_volume_path(clone)
        clone_snapshot.return_value = {}
        extend_volume.return_value = None
        self.assertIsNone(
            self.drv.create_volume_from_snapshot(clone, snapshot)
        )
        clone_payload = {'targetPath': clone_path}
        clone_snapshot.assert_called_with(snapshot_path, clone_payload)
        extend_volume.assert_called_with(clone, clone_size)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefLunMappings.list')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver._create_target_group')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver._create_target')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver._target_group_props')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver._get_host_portals')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver._get_host_group')
    @mock.patch('uuid.uuid4', fake_uuid4)
    def test_initialize_connection(self, get_host_group, get_host_portals,
                                   get_target_group_props, create_target,
                                   create_target_group, list_mappings):
        volume = fake_volume(self.ctxt)
        host_iqn = 'iqn:cinder-client'
        target_iqn = 'iqn:cinder-target'
        connector = {'initiator': host_iqn, 'multipath': True}
        host_group = 'cinder-host-group'
        target_group = 'cinder-target-group'
        target_portals = self.cfg.nexenta_iscsi_target_portals.split(',')
        get_host_group.return_value = host_group
        get_host_portals.return_value = {
            target_iqn: target_portals
        }
        list_mappings.return_value = [{
            'id': '309F9B9013CF627A00000000',
            'lun': 0,
            'hostGroup': host_group,
            'targetGroup': target_group
        }]
        get_target_group_props.return_value = {
            target_iqn: target_portals
        }
        create_target.return_value = {}
        create_target_group.return_value = {}
        result = self.drv.initialize_connection(volume, connector)
        expected = {
            'driver_volume_type': 'iscsi',
            'data': {
                'target_discovered': False,
                'encrypted': False,
                'qos_specs': None,
                'target_luns': [0] * len(target_portals),
                'access_mode': 'rw',
                'volume_id': volume['id'],
                'target_portals': target_portals,
                'target_iqns': [target_iqn] * len(target_portals)
            }
        }
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefLunMappings.delete')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefLunMappings.list')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver._get_host_group')
    def test_terminate_connection(self, get_host_group,
                                  list_mappings, delete_mapping):
        volume = fake_volume(self.ctxt)
        host_group = 'cinder-host-group'
        target_group = 'cinder-target-group'
        connector = {'initiator': 'iqn:test'}
        get_host_group.return_value = host_group
        list_mappings.return_value = [{
            'id': '309F9B9013CF627A00000000',
            'lun': 0,
            'hostGroup': host_group,
            'targetGroup': target_group
        }]
        delete_mapping.return_value = {}
        expected = {'driver_volume_type': 'iscsi', 'data': {}}
        result = self.drv.terminate_connection(volume, connector)
        self.assertEqual(expected, result)

    def test_create_export(self):
        volume = fake_volume(self.ctxt)
        connector = {'initiator': 'iqn:test'}
        self.assertIsNone(
            self.drv.create_export(self.ctxt, volume, connector)
        )

    def test_ensure_export(self):
        volume = fake_volume(self.ctxt)
        self.assertIsNone(
            self.drv.ensure_export(self.ctxt, volume)
        )

    def test_remove_export(self):
        volume = fake_volume(self.ctxt)
        self.assertIsNone(
            self.drv.remove_export(self.ctxt, volume)
        )

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumeGroups.get')
    def test_get_volume_stats(self, get_volume_group):
        available = 100
        used = 75
        get_volume_group.return_value = {
            'bytesAvailable': available * units.Gi,
            'bytesUsed': used * units.Gi
        }
        result = self.drv.get_volume_stats(True)
        payload = {'fields': 'bytesAvailable,bytesUsed'}
        get_volume_group.assert_called_with(self.drv.root_path, payload)
        self.assertEqual(self.drv._stats, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumeGroups.get')
    def test_update_volume_stats(self, get_volume_group):
        available = 8
        used = 2
        get_volume_group.return_value = {
            'bytesAvailable': available * units.Gi,
            'bytesUsed': used * units.Gi
        }
        location_info = '%(driver)s:%(host)s:%(pool)s/%(group)s' % {
            'driver': self.drv.__class__.__name__,
            'host': self.cfg.nexenta_host,
            'pool': self.cfg.nexenta_volume,
            'group': self.cfg.nexenta_volume_group,
        }
        expected = {
            'vendor_name': 'Nexenta',
            'dedup': self.cfg.nexenta_dataset_dedup,
            'compression': self.cfg.nexenta_dataset_compression,
            'description': self.cfg.nexenta_dataset_description,
            'driver_version': self.drv.VERSION,
            'storage_protocol': 'iSCSI',
            'sparsed_volumes': self.cfg.nexenta_sparse,
            'total_capacity_gb': used + available,
            'free_capacity_gb': available,
            'reserved_percentage': self.cfg.reserved_percentage,
            'QoS_support': False,
            'multiattach': True,
            'consistencygroup_support': True,
            'consistent_group_snapshot_enabled': True,
            'volume_backend_name': self.cfg.volume_backend_name,
            'location_info': location_info,
            'iscsi_target_portal_port': (
                self.cfg.nexenta_iscsi_target_portal_port),
            'nef_url': self.cfg.nexenta_rest_address,
            'nef_port': self.cfg.nexenta_rest_port
        }
        self.assertIsNone(self.drv._update_volume_stats())
        self.assertEqual(expected, self.drv._stats)

    def test__get_volume_path(self):
        volume = fake_volume(self.ctxt)
        result = self.drv._get_volume_path(volume)
        expected = '%s/%s/%s' % (self.cfg.nexenta_volume,
                                 self.cfg.nexenta_volume_group,
                                 volume['name'])
        self.assertEqual(expected, result)

    def test__get_snapshot_path(self):
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        result = self.drv._get_snapshot_path(snapshot)
        expected = '%s/%s/%s@%s' % (self.cfg.nexenta_volume,
                                    self.cfg.nexenta_volume_group,
                                    snapshot['volume_name'],
                                    snapshot['name'])
        self.assertEqual(expected, result)

    def test__get_target_group_name(self):
        target_iqn = '%s-test' % self.cfg.nexenta_target_prefix
        result = self.drv._get_target_group_name(target_iqn)
        expected = '%s-test' % self.cfg.nexenta_target_group_prefix
        self.assertEqual(expected, result)

    def test__get_target_name(self):
        target_group = '%s-test' % self.cfg.nexenta_target_group_prefix
        result = self.drv._get_target_name(target_group)
        expected = '%s-test' % self.cfg.nexenta_target_prefix
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefNetAddresses.list')
    def test__get_host_addresses(self, list_addresses):
        expected = ['1.1.1.1', '2.2.2.2', '3.3.3.3']
        return_value = []
        for address in expected:
            return_value.append({
                'addressType': 'static',
                'address': '%s/24' % address
            })
        list_addresses.return_value = return_value
        result = self.drv._get_host_addresses()
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver._get_host_addresses')
    def test__get_host_portals(self, list_addresses):
        list_addresses.return_value = ['1.1.1.1', '2.2.2.2', '3.3.3.3']
        expected = ['1.1.1.1:3260', '2.2.2.2:3260']
        result = self.drv._get_host_portals()
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefTargets.list')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefTargetsGroups.list')
    def test__target_group_props(self, list_target_groups, list_targets):
        host_portals = ['1.1.1.1:3260', '2.2.2.2:3260']
        target_group = 'cinder-test'
        list_target_groups.return_value = [{
            'name': target_group,
            'members': [
                'iqn:cinder-test'
            ]
        }]
        list_targets.return_value = [{
            'name': 'iqn:cinder-test',
            'portals': [
                {
                    'address': '1.1.1.1',
                    'port': 3260
                },
                {
                    'address': '2.2.2.2',
                    'port': 3260
                }
            ]
        }]
        expected = {'iqn:cinder-test': host_portals}
        result = self.drv._target_group_props(target_group, host_portals)
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefTargetsGroups.create')
    def test__create_target_group(self, create_target_group):
        name = 'name'
        members = ['a', 'b', 'c']
        create_target_group.return_value = {}
        self.assertIsNone(self.drv._create_target_group(name, members))
        payload = {'name': name, 'members': members}
        create_target_group.assert_called_with(payload)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefTargetsGroups.set')
    def test__update_target_group(self, update_target_group):
        name = 'name'
        members = ['a', 'b', 'c']
        update_target_group.return_value = {}
        self.assertIsNone(self.drv._update_target_group(name, members))
        payload = {'members': members}
        update_target_group.assert_called_with(name, payload)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefLunMappings.delete')
    def test__delete_lun_mapping(self, delete_mapping):
        name = 'name'
        delete_mapping.return_value = {}
        self.assertIsNone(self.drv._delete_lun_mapping(name))
        delete_mapping.assert_called_with(name)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefTargets.create')
    def test__create_target(self, create_target):
        name = 'name'
        portals = ['1.1.1.1:3260', '2.2.2.2:3260']
        create_target.return_value = {}
        self.assertIsNone(self.drv._create_target(name, portals))
        payload = {
            'name': name,
            'portals': [
                {
                    'address': '1.1.1.1',
                    'port': 3260
                },
                {
                    'address': '2.2.2.2',
                    'port': 3260
                }
            ]
        }
        create_target.assert_called_with(payload)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefHostGroups.list')
    def test__get_host_group(self, get_hostgroup):
        member = 'member1'
        get_hostgroup.return_value = [
            {
                'name': 'name1',
                'members': [
                    'member1',
                    'member2',
                    'member3'
                ]
            },
            {
                'name': 'name2',
                'members': [
                    'member4',
                    'member5',
                    'member6'
                ]
            }
        ]
        expected = 'name1'
        result = self.drv._get_host_group(member)
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefHostGroups.create')
    def test__create_host_group(self, create_host_group):
        name = 'name'
        members = ['a', 'b', 'c']
        create_host_group.return_value = {}
        self.assertIsNone(self.drv._create_host_group(name, members))
        payload = {'name': name, 'members': members}
        create_host_group.assert_called_with(payload)

    def test__s2d(self):
        portals = ['1.1.1.1:3260', '2.2.2.2:3260']
        expected = [
            {
                'address': '1.1.1.1',
                'port': 3260
            },
            {
                'address': '2.2.2.2',
                'port': 3260
            }
        ]
        result = self.drv._s2d(portals)
        self.assertEqual(expected, result)

    def test__d2s(self):
        portals = [
            {
                'address': '1.1.1.1',
                'port': 3260
            },
            {
                'address': '2.2.2.2',
                'port': 3260
            }
        ]
        expected = ['1.1.1.1:3260', '2.2.2.2:3260']
        result = self.drv._d2s(portals)
        self.assertEqual(expected, result)

    def test_create_consistencygroup(self):
        cgroup = fake_cgroup(self.ctxt)
        result = self.drv.create_consistencygroup(self.ctxt, cgroup)
        expected = {}
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver.delete_volume')
    def test_delete_consistencygroup(self, delete_volume):
        cgroup = fake_cgroup(self.ctxt)
        volume1 = fake_volume(self.ctxt)
        volume2_spec = {'id': fake.VOLUME2_ID}
        volume2 = fake_volume(self.ctxt, **volume2_spec)
        volumes = [volume1, volume2]
        delete_volume.return_value = {}
        result = self.drv.delete_consistencygroup(self.ctxt,
                                                  cgroup,
                                                  volumes)
        expected = ({}, [])
        self.assertEqual(expected, result)

    def test_update_consistencygroup(self):
        cgroup = fake_cgroup(self.ctxt)
        volume1 = fake_volume(self.ctxt)
        volume2_spec = {'id': fake.VOLUME2_ID}
        volume2 = fake_volume(self.ctxt, **volume2_spec)
        volume3_spec = {'id': fake.VOLUME3_ID}
        volume3 = fake_volume(self.ctxt, **volume3_spec)
        volume4_spec = {'id': fake.VOLUME4_ID}
        volume4 = fake_volume(self.ctxt, **volume4_spec)
        add_volumes = [volume1, volume2]
        remove_volumes = [volume3, volume4]
        result = self.drv.update_consistencygroup(self.ctxt,
                                                  cgroup,
                                                  add_volumes,
                                                  remove_volumes)
        expected = ({}, [], [])
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.delete')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.rename')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.create')
    def test_create_cgsnapshot(self, create_snapshot,
                               rename_snapshot,
                               delete_snapshot):
        cgsnapshot = fake_cgsnapshot(self.ctxt)
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        snapshots = [snapshot]
        cgsnapshot_name = (
            self.cfg.nexenta_group_snapshot_template % cgsnapshot['id'])
        cgsnapshot_path = '%s@%s' % (self.drv.root_path, cgsnapshot_name)
        snapshot_path = '%s/%s@%s' % (self.drv.root_path,
                                      snapshot['volume_name'],
                                      cgsnapshot_name)
        create_snapshot.return_value = {}
        rename_snapshot.return_value = {}
        delete_snapshot.return_value = {}
        result = self.drv.create_cgsnapshot(self.ctxt,
                                            cgsnapshot,
                                            snapshots)
        create_payload = {'path': cgsnapshot_path, 'recursive': True}
        create_snapshot.assert_called_with(create_payload)
        rename_payload = {'newName': snapshot['name']}
        rename_snapshot.assert_called_with(snapshot_path, rename_payload)
        delete_payload = {'defer': True, 'recursive': True}
        delete_snapshot.assert_called_with(cgsnapshot_path, delete_payload)
        expected = ({}, [])
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver.delete_snapshot')
    def test_delete_cgsnapshot(self, delete_snapshot):
        cgsnapshot = fake_cgsnapshot(self.ctxt)
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        snapshots = [snapshot]
        delete_snapshot.return_value = {}
        result = self.drv.delete_cgsnapshot(self.ctxt,
                                            cgsnapshot,
                                            snapshots)
        delete_snapshot.assert_called_with(snapshot)
        expected = ({}, [])
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver.create_volume_from_snapshot')
    def test_create_consistencygroup_from_src_snapshots(self, create_volume):
        cgroup = fake_cgroup(self.ctxt)
        cgsnapshot = fake_cgsnapshot(self.ctxt)
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        snapshots = [snapshot]
        clone_spec = {'id': fake.VOLUME2_ID}
        clone = fake_volume(self.ctxt, **clone_spec)
        clones = [clone]
        create_volume.return_value = {}
        result = self.drv.create_consistencygroup_from_src(self.ctxt, cgroup,
                                                           clones, cgsnapshot,
                                                           snapshots, None,
                                                           None)
        create_volume.assert_called_with(clone, snapshot)
        expected = ({}, [])
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.delete')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver.create_volume_from_snapshot')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.create')
    def test_create_consistencygroup_from_src_volumes(self,
                                                      create_snapshot,
                                                      create_volume,
                                                      delete_snapshot):
        src_cgroup = fake_cgroup(self.ctxt)
        dst_cgroup_spec = {'id': fake.CONSISTENCY_GROUP2_ID}
        dst_cgroup = fake_cgroup(self.ctxt, **dst_cgroup_spec)
        src_volume = fake_volume(self.ctxt)
        src_volumes = [src_volume]
        dst_volume_spec = {'id': fake.VOLUME2_ID}
        dst_volume = fake_volume(self.ctxt, **dst_volume_spec)
        dst_volumes = [dst_volume]
        create_snapshot.return_value = {}
        create_volume.return_value = {}
        delete_snapshot.return_value = {}
        result = self.drv.create_consistencygroup_from_src(self.ctxt,
                                                           dst_cgroup,
                                                           dst_volumes,
                                                           None, None,
                                                           src_cgroup,
                                                           src_volumes)
        snapshot_name = (
            self.cfg.nexenta_origin_snapshot_template % dst_cgroup['id'])
        snapshot_path = '%s@%s' % (self.drv.root_path, snapshot_name)
        create_payload = {'path': snapshot_path, 'recursive': True}
        create_snapshot.assert_called_with(create_payload)
        snapshot = {
            'name': snapshot_name,
            'volume_id': src_volume['id'],
            'volume_name': src_volume['name'],
            'volume_size': src_volume['size']
        }
        create_volume.assert_called_with(dst_volume, snapshot)
        delete_payload = {'defer': True, 'recursive': True}
        delete_snapshot.assert_called_with(snapshot_path, delete_payload)
        expected = ({}, [])
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumes.list')
    def test__get_existing_volume(self, list_volumes):
        volume = fake_volume(self.ctxt)
        parent = self.drv.root_path
        name = volume['name']
        size = volume['size']
        path = self.drv._get_volume_path(volume)
        list_volumes.return_value = [{
            'name': name,
            'path': path,
            'volumeSize': size * units.Gi
        }]
        result = self.drv._get_existing_volume({'source-name': name})
        payload = {
            'parent': parent,
            'fields': 'name,path,volumeSize',
            'name': name
        }
        list_volumes.assert_called_with(payload)
        expected = {
            'name': name,
            'path': path,
            'size': size
        }
        self.assertEqual(expected, result)

    def test__check_already_managed_snapshot(self):
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        result = self.drv._check_already_managed_snapshot(snapshot)
        expected = False
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.list')
    def test__get_existing_snapshot(self, list_snapshots):
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        name = snapshot['name']
        path = self.drv._get_snapshot_path(snapshot)
        parent = self.drv._get_volume_path(volume)
        list_snapshots.return_value = [{
            'name': name,
            'path': path
        }]
        payload = {'source-name': name}
        result = self.drv._get_existing_snapshot(snapshot, payload)
        payload = {
            'parent': parent,
            'fields': 'name,path',
            'recursive': False,
            'name': name
        }
        list_snapshots.assert_called_with(payload)
        expected = {
            'name': name,
            'path': path,
            'volume_name': volume['name'],
            'volume_size': volume['size']
        }
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumes.rename')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefLunMappings.list')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver._get_existing_volume')
    def test_manage_existing(self, get_existing_volume,
                             list_mappings, rename_volume):
        existing_volume = fake_volume(self.ctxt)
        manage_volume_spec = {'id': fake.VOLUME2_ID}
        manage_volume = fake_volume(self.ctxt, **manage_volume_spec)
        existing_name = existing_volume['name']
        existing_path = self.drv._get_volume_path(existing_volume)
        existing_size = existing_volume['size']
        manage_path = self.drv._get_volume_path(manage_volume)
        get_existing_volume.return_value = {
            'name': existing_name,
            'path': existing_path,
            'size': existing_size
        }
        list_mappings.return_value = []
        payload = {'source-name': existing_name}
        self.assertIsNone(self.drv.manage_existing(manage_volume, payload))
        get_existing_volume.assert_called_with(payload)
        payload = {'volume': existing_path}
        list_mappings.assert_called_with(payload)
        payload = {'newPath': manage_path}
        rename_volume.assert_called_with(existing_path, payload)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.iscsi.'
                'NexentaISCSIDriver._get_existing_volume')
    def test_manage_existing_get_size(self, get_volume):
        volume = fake_volume(self.ctxt)
        name = volume['name']
        size = volume['size']
        path = self.drv._get_volume_path(volume)
        get_volume.return_value = {
            'name': name,
            'path': path,
            'size': size
        }
        payload = {'source-name': name}
        result = self.drv.manage_existing_get_size(volume, payload)
        expected = size
        self.assertEqual(expected, result)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefVolumes.list')
    def test_get_manageable_volumes(self, list_volumes):
        volume = fake_volume(self.ctxt)
        volumes = [volume]
        name = volume['name']
        size = volume['size']
        path = self.drv._get_volume_path(volume)
        guid = 12345
        parent = self.drv.root_path
        list_volumes.return_value = [{
            'name': name,
            'path': path,
            'guid': guid,
            'volumeSize': size * units.Gi
        }]
        result = self.drv.get_manageable_volumes(volumes, None, 1,
                                                 0, 'size', 'asc')
        payload = {
            'parent': parent,
            'fields': 'name,guid,path,volumeSize',
            'recursive': False
        }
        list_volumes.assert_called_with(payload)
        expected = [{
            'cinder_id': volume['id'],
            'extra_info': None,
            'reason_not_safe': 'Volume already managed',
            'reference': {
                'source-guid': guid,
                'source-name': volume['name']
            },
            'safe_to_manage': False,
            'size': volume['size']
        }]
        self.assertEqual(expected, result)

    def test_unmanage(self):
        volume = fake_volume(self.ctxt)
        self.assertIsNone(self.drv.unmanage(volume))

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.rename')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver._get_existing_snapshot')
    def test_manage_existing_snapshot(self, get_existing_snapshot,
                                      rename_snapshot):
        volume = fake_volume(self.ctxt)
        existing_snapshot = fake_snapshot(self.ctxt)
        existing_snapshot.volume = volume
        manage_snapshot_spec = {'id': fake.SNAPSHOT2_ID}
        manage_snapshot = fake_snapshot(self.ctxt, **manage_snapshot_spec)
        manage_snapshot.volume = volume
        existing_name = existing_snapshot['name']
        manage_name = manage_snapshot['name']
        volume_name = volume['name']
        volume_size = volume['size']
        existing_path = self.drv._get_snapshot_path(existing_snapshot)
        get_existing_snapshot.return_value = {
            'name': existing_name,
            'path': existing_path,
            'volume_name': volume_name,
            'volume_size': volume_size
        }
        rename_snapshot.return_value = {}
        payload = {'source-name': existing_name}
        self.assertIsNone(
            self.drv.manage_existing_snapshot(manage_snapshot, payload)
        )
        get_existing_snapshot.assert_called_with(manage_snapshot, payload)
        payload = {'newName': manage_name}
        rename_snapshot.assert_called_with(existing_path, payload)

    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'iscsi.NexentaISCSIDriver._get_existing_snapshot')
    def test_manage_existing_snapshot_get_size(self, get_snapshot):
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        snapshot_name = snapshot['name']
        volume_name = volume['name']
        volume_size = volume['size']
        snapshot_path = self.drv._get_snapshot_path(snapshot)
        get_snapshot.return_value = {
            'name': snapshot_name,
            'path': snapshot_path,
            'volume_name': volume_name,
            'volume_size': volume_size
        }
        payload = {'source-name': snapshot_name}
        result = self.drv.manage_existing_snapshot_get_size(volume, payload)
        expected = volume['size']
        self.assertEqual(expected, result)

    @mock.patch('cinder.objects.VolumeList.get_all_by_host')
    @mock.patch('cinder.volume.drivers.nexenta.ns5.'
                'jsonrpc.NefSnapshots.list')
    def test_get_manageable_snapshots(self, list_snapshots, list_volumes):
        volume = fake_volume(self.ctxt)
        volumes = [volume]
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        snapshots = [snapshot]
        guid = 12345
        name = snapshot['name']
        path = self.drv._get_snapshot_path(snapshot)
        parent = self.drv._get_volume_path(volume)
        list_snapshots.return_value = [{
            'name': name,
            'path': path,
            'guid': guid,
            'parent': parent,
            'hprService': '',
            'snaplistId': ''
        }]
        list_volumes.return_value = volumes
        result = self.drv.get_manageable_snapshots(snapshots, None, 1,
                                                   0, 'size', 'asc')
        payload = {
            'parent': self.drv.root_path,
            'fields': 'name,guid,path,parent,hprService,snaplistId',
            'recursive': True
        }
        list_snapshots.assert_called_with(payload)
        expected = [{
            'cinder_id': snapshot['id'],
            'extra_info': None,
            'reason_not_safe': 'Snapshot already managed',
            'source_reference': {
                'name': volume['name']
            },
            'reference': {
                'source-guid': guid,
                'source-name': snapshot['name']
            },
            'safe_to_manage': False,
            'size': volume['size']
        }]
        self.assertEqual(expected, result)

    def test_unmanage_snapshot(self):
        volume = fake_volume(self.ctxt)
        snapshot = fake_snapshot(self.ctxt)
        snapshot.volume = volume
        self.assertIsNone(self.drv.unmanage_snapshot(snapshot))



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
import six

from cinder.scheduler.evaluator import evaluator
from cinder.scheduler import filters


LOG = logging.getLogger(__name__)


class DriverFilter(filters.BaseBackendFilter):
    """DriverFilter filters backend based on a 'filter function' and metrics.

    DriverFilter filters based on volume backend's provided 'filter function'
    and metrics.
    """

    def backend_passes(self, backend_state, filter_properties):
        """Determines if a backend has a passing filter_function or not."""
        stats = self._generate_stats(backend_state, filter_properties)

        LOG.debug("Checking backend '%s'",
                  stats['backend_stats']['backend_id'])
        result = self._check_filter_function(stats)
        LOG.debug("Result: %s", result)
        LOG.debug("Done checking backend '%s'",
                  stats['backend_stats']['backend_id'])

        return result

    def _check_filter_function(self, stats):
        """Checks if a volume passes a backend's filter function.

           Returns a tuple in the format (filter_passing, filter_invalid).
           Both values are booleans.
        """
        if stats['filter_function'] is None:
            LOG.debug("Filter function not set :: passing backend")
            return True

        try:
            filter_result = self._run_evaluator(stats['filter_function'],
                                                stats)
        except Exception as ex:
            # Warn the admin for now that there is an error in the
            # filter function.
            LOG.warning("Error in filtering function "
                        "'%(function)s' : '%(error)s' :: failing backend",
                        {'function': stats['filter_function'],
                         'error': ex, })
            return False

        return filter_result

    def _run_evaluator(self, func, stats):
        """Evaluates a given function using the provided available stats."""
        backend_stats = stats['backend_stats']
        backend_caps = stats['backend_caps']
        extra_specs = stats['extra_specs']
        qos_specs = stats['qos_specs']
        volume_stats = stats['volume_stats']

        LOG.debug('Running evaluator: extra_specs: %(extra)s\n'
                  'stats: %(stats)s\n'
                  'capabilities: %(capabilities)s\n'
                  'volume: %(volume)s\n'
                  'qos: %(qos)s', {'extra': extra_specs,
                                   'stats': backend_stats,
                                   'capabilities': backend_caps,
                                   'volume': volume_stats,
                                   'qos': qos_specs})

        result = evaluator.evaluate(
            func,
            extra=extra_specs,
            stats=backend_stats,
            capabilities=backend_caps,
            volume=volume_stats,
            qos=qos_specs)

        return result

    def _generate_stats(self, backend_state, filter_properties):
        """Generates statistics from backend and volume data."""

        backend_stats = {
            'host': backend_state.host,
            'cluster_name': backend_state.cluster_name,
            'backend_id': backend_state.backend_id,
            'volume_backend_name': backend_state.volume_backend_name,
            'vendor_name': backend_state.vendor_name,
            'driver_version': backend_state.driver_version,
            'storage_protocol': backend_state.storage_protocol,
            'QoS_support': backend_state.QoS_support,
            'total_capacity_gb': backend_state.total_capacity_gb,
            'allocated_capacity_gb': backend_state.allocated_capacity_gb,
            'free_capacity_gb': backend_state.free_capacity_gb,
            'reserved_percentage': backend_state.reserved_percentage,
            'updated': backend_state.updated,
        }

        backend_caps = backend_state.capabilities

        filter_function = None

        if ('filter_function' in backend_caps and
                backend_caps['filter_function'] is not None):
            filter_function = six.text_type(backend_caps['filter_function'])

        qos_specs = filter_properties.get('qos_specs', {})

        volume_type = filter_properties.get('volume_type', {})
        extra_specs = volume_type.get('extra_specs', {})

        request_spec = filter_properties.get('request_spec', {})
        volume_stats = request_spec.get('volume_properties', {})

        stats = {
            'backend_stats': backend_stats,
            'backend_caps': backend_caps,
            'extra_specs': extra_specs,
            'qos_specs': qos_specs,
            'volume_stats': volume_stats,
            'volume_type': volume_type,
            'filter_function': filter_function,
        }

        return stats
